.cpu "65c02"
; Hello world
;
; Assembly program for x16, using 64tass


; Basic header
; 10 SYS 2064

*=$0801
.byte $0C,$08,$0A,$00,$9E,$20,$32,$30,$36,$34,$00,$00,$00
*=$0810

ADDRx_L=$9F20
ADDRx_M=$9F21
ADDRx_H=$9F22
DATA0=$9F23
L0_CONFIG=$9F2D
L0_TILEBASE=$9F2F
L0_HSCROLL_H=$9F31
SPI_DATA=$9F3E
SPI_CTRL=$9F3F
loop_ctr1 = $23
AUDIO_CTRL := $9F3B
AUDIO_FIFO := $9F3D
AUDIO_RATE := $9F3C
VERA_ISR  := $9F27

EXTAPI = $FEAB
GETIN = $FFE4
SETNAM = $FFBD
SETLFS = $FFBA
OPEN = $FFC0
CHKIN = $FFC6
CLOSE = $FFC3
CLRCHN = $FFCC
MACPTR = $FF44

; Kernal functions
CHROUT=$FFD2     ; Kernal: CHROUT outputs a character to screen
RDTIM=$FFDE     ; Kernal: Read jiffy clock
screen_mode=$FF5F


; Control characters
CTRL_RETURN=$0D  ; New line
CTRL_CLS=$93     ; Clear screen

r0=2
r0L=2
r0H=3
r1=4
r1L=4
r1H=5
r2=6
r2L=6
r2H=7
r3=8
r3L=8
r3H=9
r4=10
r4L=10
r4H=11
r10L = 22
r10H = 23
r11L = 24
r11H = 25



STARTTIME=$30 ; 30-32, BE
ENDTIME=$33 ; 33-35, BE
ELAPSEDTIME=$36 ; 36
CNT=$40 ; 40-41, LE
FRAME_REMAINING=$50 ; 50-51, LE
ACTIVE_BUFFER=$52 ; bit 7

SPI_CTRL_AUTOTX        = $04


; Absolute addresses inside R48
sector_buffer=$B9C2 ; RAM bank 0. This is NOT fixed, R48 is required.
cmd_idx=$BBC2
sector_lba=$BBC3 ; RAM bank 0. This is NOT fixed, R48 is required.
cmd_crc=$BBC7
sd_fast=$BBC8 ; RAM bank 0. This is NOT fixed, R48 is required. Set 
sdcard_read_sector=$E385 ; ROM bank 3. This is NOT fixed, R48 is required.
send_cmd=$E268 ; ROM bank 3. This is NOT fixed, R48 is required.
spi_read=$E251
deselect=$E230


    ;!byte $db ; stop in emulator

    lda #CTRL_CLS
    jsr CHROUT
    
    
; Print string "HELLO WORLD!"
    ldx #<STR_HELLO_WORLD
    ldy #>STR_HELLO_WORLD
    jsr printStringNewLine


; Test 0: Require exactly ROM version 48
    ldx 1
    stz 1
    lda $FF80
    stx 1
    cmp #48
    beq rtest1

    ldx #<STR_INCOMPATIBLE_ROM
    ldy #>STR_INCOMPATIBLE_ROM
    jsr printStringNewLine
    rts
rtest1:
    ldx #<STR_ROM_OK
    ldy #>STR_ROM_OK
    jsr printStringNewLine


    ;jsr macptrDemo


; Test 1: Fetch basic SD card properties
; Get SD card type (MMC/SD/SDHC/SDXC/..)
; Read CID/CSD/OCR/voltage range
; Read first sector and get file system properties

; Test 2: Read blocks from big file using MACPTR (skip if file not found)

; Test 3: Read random block using sdcard_read_sector
    ;!byte $db
    ; This works
    ;jsr readSingleBlockToRam
    
    ; Measure the time to call sdcard_read_sector N times
    ;jsr readSingleBlockToRam_measure_time_N

; Test 4: Read random blocks using custom sdcard_read (fast readout)

; Test 5: Read sequential blocks using READ_MULTIPLE_BLOCK

    ;jsr readSeqToVram

    jsr streamVideo

; Return to basic
	rts

debugAXY:
    ; Output: FF 00 FF 00 FF 00 FF 00   [A] [X] [Y] 00 EE 00 EE 00
    phy
    phy
    ldy #$FF
    
    sty DATA0
    stz DATA0
    sty DATA0
    stz DATA0
    sty DATA0
    stz DATA0
    sty DATA0
    stz DATA0
    
    sta DATA0
    stx DATA0
    ply
    sty DATA0
    ldy #$EE

    stz DATA0
    sty DATA0
    stz DATA0
    sty DATA0
    stz DATA0
    
    ply
    rts
    



error:
    ; error code in X
    ; A values:
    ; 22: Block received (not error)
    ; 33: Block received (not error)
    lda #$33
    jsr debugAXY ; A: Location, X/Y: data
    brk
    nop
    nop
    .byte $db ; do not proceed
    




clearAndPrepareVram:
    stz ADDRx_L
    stz ADDRx_M
    lda #$10 ; addr incr = +1
    sta ADDRx_H
    
    ldx #0
l1:
    ldy #0
l2:
    stz DATA0
    dey
    bne l2
    dex
    bne l1

    stz ADDRx_L
    stz ADDRx_M
    ;lda #$10 ; addr incr = +1
    lda #0 ; for reading 1000 sectors
    sta ADDRx_H


    rts




; Modded sdcard_read_sector, using parts of sdcard_read_sector
readToVram:
    
	; Send READ_SINGLE_BLOCK command
	lda #($40 | 17)
	sta cmd_idx
	lda #1
	sta cmd_crc
	jsr send_cmd

	; Wait for start of data packet
	ldx #0
l3:	ldy #0
l4:	jsr spi_read
	cmp #$FE
	beq readToVram_start
	dey
	bne l4
	dex
	bne l3

	; Timeout error
	jsr deselect
	clc
	rts

readToVram_start:
	; Enable auto-tx mode
	lda SPI_CTRL
	ora #SPI_CTRL_AUTOTX
	sta SPI_CTRL

	; Start first read transfer
	lda SPI_DATA			; Auto-tx
	ldy #0				; 2

    ; Read directly to vera
    jsr copyFromSpiToVeraData0

	; Disable auto-tx mode
	lda SPI_CTRL
	;and #(SPI_CTRL_AUTOTX ^ $FF)
    and #$FB ; ~$04
	sta SPI_CTRL

	; Next read is now already done (first CRC byte), read second CRC byte
	jsr spi_read

	; Success
	jsr deselect
	sec
	rts


    


CLEARGRAPH: ; courtesy of Tony 3068
        stz $9F20
        stz $9F21
        lda #%00010000
        sta $9F22
        lda #0
        ldx #25
Cloop1:  stz $9F23
         stz $9F23
         stz $9F23
         stz $9F23
         stz $9F23
         stz $9F23
         stz $9F23
         stz $9F23
         stz $9F23
         stz $9F23
         stz $9F23
         stz $9F23
        ina
        bne Cloop1
        dex
        bne Cloop1
      rts



prepareVeraForVideo:

    ; Start with video mode $80
    lda #$80
    clc
    jsr screen_mode

    lda L0_CONFIG
    ; Color depth is bit 0 and 1
    ; Default 11 (8)
    ; Change to 10 (4)
    and #$FE
    sta L0_CONFIG


    jsr CLEARGRAPH

    ; First iteration: Make 09800 visible and draw to 0

    stz ADDRx_L
    stz ADDRx_M
    lda #$10 ; addr incr = +1
    sta ADDRx_H

    lda #$76 ; Point to 0x09800
    sta L0_TILEBASE

    lda #$80
    sta ACTIVE_BUFFER

    rts



; Data is inside cur_context (struct context)








dummyBuff:
    .byte 0,0,0,0,0

currContextLba = $B7E1

macptrDemo:

        ;.byte $db

        lda #filename_len
        ldx #<filename
        ldy #>filename
        jsr SETNAM
        lda #1
        ldx #8
        ldy #2
        jsr SETLFS
        jsr OPEN
        
        ; File sector is stored inside the struct "cur_context", at "lba".
        ; Absolute offset, R48: $B7E1, bank 0. 4 byte little endian.
        
        ldy 0
        stz 0
        lda currContextLba
        sta r0L
        lda currContextLba+1
        sta r0H
        lda currContextLba+2
        sta r1L
        lda currContextLba+3
        sta r1H
        
        sty 0
        
        lda #1
        jsr CLOSE
        
        

        brk
        
        jsr OPEN

        
        ldx #1
        jsr CHKIN
        
        lda #5
        ldx #<dummyBuff
        ldy #>dummyBuff
        jsr MACPTR
        
        

    ; cleanup
        jsr CLRCHN
        lda #1
        jmp CLOSE



; Get the first sector of file "MJS.SPV"
getFileSector:
    lda #filename_len
    ldx #<filename
    ldy #>filename
    jsr SETNAM
    lda #1
    ldx #8
    ldy #2
    jsr SETLFS
    jsr OPEN
    
    ; File sector is stored inside the struct "cur_context", at "lba".
    ; Absolute offset, R48: $B7E1, bank 0. 4 byte little endian.
    
    ldy 0
    stz 0
    lda currContextLba
    sta r0L
    lda currContextLba+1
    sta r0H
    lda currContextLba+2
    sta r1L
    lda currContextLba+3
    sta r1H
    
    sty 0
    
    lda #1
    jmp CLOSE



filename:
    ;!byte "VIDEO.SPV"
    .text "SHADOW16.SPV"
filename_len = *-filename

openAudioFile:
 lda 1
 pha
 stz 1
    lda #afile_length
    ldx #<afilename
    ldy #>afilename
    jsr SETNAM
    lda #6
    ldx #8
    ldy #6
    jsr SETLFS
    jsr OPEN
 pla
 sta 1
 rts

afilename:
   .text "SHADOW16.SVA,S,R"
afile_length = *-afilename


CopyZeroPal:
    ;asl
    ;asl
    ;asl
    ;asl
    sta ADDRx_L
    lda #$FB
    sta ADDRx_M
    lda #%00010001
    sta ADDRx_H
    ldx $9F23
    ldy $9F23
    stz ADDRx_L
    lda #$FA
    sta ADDRx_M
    stx $9F23
    sty $9F23
   rts

streamVideo:
     jsr openAudioFile
    ; Use sdcard_read_sector in rom bank 3 (FAT32/SD)
    ; This function assumes rom bank 0
    lda 0
    pha
    lda 1
    pha
    
    stz 0 ; RAM bank 0 (kernal)
    

    


    ; Set fast tx. Restore to slow at the end.
    lda #$40
    sta sd_fast
    


    stz AUDIO_RATE
    lda #%10000000
    sta AUDIO_CTRL    ; reset the Audio FIFO
    lda #$0E          ; 8 bit mono volume 14
    sta AUDIO_CTRL
    stz AUDIO_RATE
    jsr LoadAudioFrame
    
    jsr prepareVeraForVideo

    ; Get first sector of "SHADOWTRY2.SPV" into r0/r1
    jsr getFileSector

    lda #3
    sta 1 ; ROM bank 3 (FAT32)




    lda r0L
    sta sector_lba
    lda r0H
    sta sector_lba+1
    lda r1L
    sta sector_lba+2
    lda r1H
    sta sector_lba+3


    ; Add 1 to sector
    inc sector_lba+0
    bne endIncSector
    inc sector_lba+1
    bne endIncSector
    inc sector_lba+2
    bne endIncSector
    inc sector_lba+3
endIncSector:



    ; There are 12328 frames
    FRAME_COUNT=2834
    
    lda #<FRAME_COUNT
    sta FRAME_REMAINING
    lda #((>FRAME_COUNT) + 1) ; add one due to dec+beq
    sta FRAME_REMAINING+1

    ; Set callback (r3): Read to DATA0
    lda #<copyFromSpiToVeraData0
    sta r3L
    lda #>copyFromSpiToVeraData0
    sta r3H

    lda #84
    sta AUDIO_RATE
    
next_frame:

    ; Load 75 sectors to DATA0 (frame)
    lda #75
    ;sta CNT
    ;stz CNT+1
    sta r2L ; sector count
    stz r2H

    ;jsr readToVram
    ;jsr readToVramMulti
    ;jsr readToVramMultiFake
    ;jsr readMultiBlockUsingCallback
    jsr readMultiBlockToVera0
    bcs noerr
    jmp error
noerr:
    ; Add 75 to sector
    lda sector_lba
    ;lda r0L
    clc
    adc #75
    sta sector_lba
    ;sta r0L
    bcc endincsect2
    inc sector_lba+1
    ;inc r0H
    bne endincsect2
    inc sector_lba+2
    ;inc r1L
    bne endincsect2
    inc sector_lba+3
    ;inc r1H
endincsect2:


    ; Point to palette, at 1FA00
    stz ADDRx_L
    lda #$FA
    sta ADDRx_M
    lda #($10 | 1) ; addr incr = +1
    sta ADDRx_H


    ; Load 1 sectors to DATA0 (palette)
    lda #1
    ;sta CNT
    ;stz CNT+1
    sta r2L
    stz r2H

    ;jsr readToVram
    ;jsr readToVramMulti
    ;jsr readToVramMultiFake
    ;jsr readMultiBlockUsingCallback
    jsr readMultiBlockToVera0
    bcs skiperr2
    jmp error
skiperr2:

    
    ; Add 1 to sector
    inc sector_lba+0
    ;inc r0L
    bne endincsect3
    inc sector_lba+1
    ;inc r0H
    bne endincsect3
    inc sector_lba+2
    ;inc r1L
    bne endincsect3
    inc sector_lba+3
    ;inc r1H
endincsect3:


    ; Wait for next vblank
  .byte $CB  ;  wai
    
    
    ; Buffers:
    ; 320x240 4bpp = 38400 bytes = 0x9600 bytes
    ; 00000-095FF: buffer 0
    ; 09800-12DFF: buffer 1
    
    
    ; Flip buffers
    lda ACTIVE_BUFFER
    eor #$80
    sta ACTIVE_BUFFER
    
    ;bpl flipTo1
    bmi flipTo1
flipTo0:
    ; 7-2: base address: upper 6 bits of address to bitmap (01001011) -> (010011 -> 0 1001 1000 0000)
    ; 1: tile height
    ; 0: tile width (0: 320 px)
    stz L0_TILEBASE
    
    ; Set palette index 14 (frame 0,2,4,...)
    ;lda #14
    lda #$0E
    sta L0_HSCROLL_H
    
    ;lda #$E0
    ;jsr CopyZeroPal
    
    ; Set address to 9800
    stz ADDRx_L
    lda #$98
    sta ADDRx_M
    lda #$10 ; addr incr = +1
    sta ADDRx_H
    
    bra skipflip1

flipTo1:
    ;lda #$76 ; Point to 0x09800 -> incorrect!
    lda #$4C ; Point to 0x9800
    sta L0_TILEBASE

    ; Set palette index 15 (frame 1,3,5,...)
    ;lda #15
    lda #$0F
    sta L0_HSCROLL_H

    ;lda #$F0
    ;jsr CopyZeroPal


    ; Set address to 0
    stz ADDRx_L
    stz ADDRx_M
    lda #$10 ; addr incr = +1
    sta ADDRx_H

skipflip1:

    dec FRAME_REMAINING
    bne endFrameDec
    dec FRAME_REMAINING+1
    bne endFrameDec
    
    ; last frame, "brk" for now
    jmp DoneStream
    .byte $db
    
endFrameDec:
          lda VERA_ISR
          and #$08
          beq endFrameDec   ; wait for AFLOW

     jsr LoadAudioFrame

    jmp next_frame

DoneStream:
    ; restore rom
    stz AUDIO_RATE ; Stop Audio
    pla
    sta 1
    lda #6
    jmp CLOSE




; Print zero-terminated string
; X: String pointer, low
; Y: String pointer, high
; Uses 3 bytes of stack to avoid scratching
printString:
    pha
    lda r0L
    pha
    lda r0H
    pha
    stx r0L
    sty r0H
l6:
    lda (r0)
    beq el6
    jsr CHROUT
    inc r0L
    bne l6
    inc r0H
    bra l6
el6:
    pla
    sta r0H
    pla
    sta r0L
    pla
    rts


printStringNewLine:
    jsr printString
printNewLine:
    pha
    lda #CTRL_RETURN
    jsr CHROUT
    pla
    rts




; This is available in sdcard-lib-r48:

; readMultiBlockUsingCallback
; Input:  r0-r1: Sector
;         r2:    Sector count
;         r3:    Callback function
; Output: r4:    Successfully copied sectors
;         C:     Success if carry set
; r3: address of a callback function, in low ram, which is expected to read 512 bytes from SPI_DATA using Auto-TX
; Note that this function will only work with rom R48 (at the moment)



;   !source "sdcard-lib.asm"


; SD card library for reading raw sectors
; Integrated with rom R48, requires this exact rom version


;ROM_INDEPENDENT=1



; Absolute addresses inside R48









; Interface:

; readSingleBlockToBuffer
; Input:  r0-r1: Sector
; Output: C=0: Error, C=1: Success
; This function is a simple wrapper function for sdcard_read_sector in ROM.
; The user can read the contents in ram bank 0, address $B9C2 (sector_buffer), 512 bytes.
;
; Note that this function will only work with rom R48.


; readMultiBlockUsingCallback
; Input:  r0-r1: Sector
;         r2:    Sector count
;         r3:    Callback function
; Output: r4:    Successfully copied sectors
;         C:     Success if carry set

; r3: address of a callback function, in low ram, which is expected to read 512 bytes from SPI_DATA using Auto-TX
; Note that this function will only work with rom R48 (at the moment)


readSingleBlockToBuffer:

    ; Use sdcard_read_sector in rom bank 3 (FAT32/SD)
    ; sdcard_read_sector assumes rom bank 0

    ; Preserve current ram/rom bank
    lda 0
    pha
    lda 1
    pha

    stz 0 ; RAM bank 0 (kernal)
    lda #3
    sta 1 ; ROM bank 3 (FAT32)

    ; Set sector_lba to r0/r1
    lda r0L
    sta sector_lba
    lda r0H
    sta sector_lba+1
    lda r1L
    sta sector_lba+2
    lda r1H
    sta sector_lba+3

    ; Set fast tx. Restore to slow at the end.
    lda #$40
    sta sd_fast

    jsr sdcard_read_sector

    ; Clear fast TX
    stz sd_fast

    ; restore rom/ram
    pla
    sta 1
    pla
    sta 0
    rts




; Input:  r0-r1: Sector
;         r2:    Sector count
;         r3:    Callback function
; Output: r4:    Successfully copied sectors
;         C:     Success if carry set

; r3: address of a callback function, in low ram, which is expected to read 512 bytes from SPI_DATA using Auto-TX
; Note that this function will only work with rom R48 (at the moment)

; Based on sdcard_read_sector, but using CMD18 (READ_MULTIPLE_BLOCK)
; Assumes rom bank 3 ram bank 0
; If error: Error code in X
;readMultiBlockUsingCallback:
readMultiBlockToVera0:
	; Send CMD18 (READ_MULTIPLE_BLOCK)
	lda #($40 | 18)
	sta cmd_idx
	lda #1
	sta cmd_crc
	jsr send_cmd

    stz r4L
    stz r4H

    ;lda r0L
    ;sta sector_lba
    ;lda r0H
    ;sta sector_lba+1
    ;lda r1L
    ;sta sector_lba+2
    ;lda r1H
    ;sta sector_lba+3

    ldx #1 ; (If carry clear) Error 1: Issue with sending CMD18 (READ_MULTIPLE_BLOCK)
    bcc deselectAndReturn

readMultiBlockUsingCallback_waitForNextBlock:
	; Wait for start of data packet
	 ldx #0
rm1: ldy #0
rm2: jsr spi_read
	 cmp #$FE
	 beq readMultiBlockUsingCallback_startReceiveBlock
	 dey
	 bne rm2
	 dex
	 bne rm1

    ; Timeout
    jsr stopTransmissionAndDeselect
    ldx #2 ; Error 2: Timeout while waiting for block
	clc
	rts

deselectAndReturn:
    ; If error: Ensure carry is cleared
    jmp deselect

;readCallback:
;    jmp (r3)

readMultiBlockUsingCallback_startReceiveBlock:
    lda #$05 ; autotx + select
	sta SPI_CTRL

	; Start first read transfer
	lda SPI_DATA ; Auto-tx

    ; call callback function in r3 to read 512 bytes from SPI_DATA using Auto-TX
    ;jsr (r3)
    ;jsr ($0008)
    ;jsr readCallback
    jsr copyFromSpiToVeraData0

	; Disable auto-tx mode, keep select
    lda #$01
	sta SPI_CTRL

	; Next read is now already done (first CRC byte), read second CRC byte
	jsr spi_read

    ; Block have been read.
    inc r4L
    bne sbh
    inc r4H
sbh:
    ; Have all blocks been read?
    lda r2L
    cmp r4L
    bne readMultiBlockUsingCallback_waitForNextBlock
    lda r2H
    cmp r4H
    bne readMultiBlockUsingCallback_waitForNextBlock


    ; All blocks have been read.
    ; Abort with cmd12

stopTransmissionAndDeselect:
    ; Send CMD12 (STOP_TRANSMISSION)
	lda #($40 | 12)
	sta cmd_idx
	lda #1
	sta cmd_crc
	jsr send_cmd
    ldx #3 ; (If carry clear) Error 3: Issue with sending CMD12 (STOP_TRANSMISSION)

    bra deselectAndReturn







; Unrolled copy loop: Read from SPI to VERA DATA0, 512 times
copyFromSpiToVeraData0:

         ldx #4
    ; need to expand this 64 times
spiloop:
     .rept 128
         lda SPI_DATA
         sta DATA0
     .endrept
         dex
         beq copyVeraDone
         jmp spiloop
copyVeraDone:
    rts

LoadAudioFrame:
    lda sector_lba
    sta r10L
    lda sector_lba+1
    sta r10H
    lda sector_lba+2
    sta r11L
    lda sector_lba+3
    sta r11H

  lda 1
  pha
  stz 1

audlfn1:
          ldx #6
          jsr CHKIN
abuff255: lda #7
          sta loop_ctr1
aloop1:   ldx #<AUDIO_FIFO
          ldy #>AUDIO_FIFO
          lda #255
          sec
          jsr MACPTR
          dec loop_ctr1
          bne aloop1
          ldx #<AUDIO_FIFO
          ldy #>AUDIO_FIFO
abuffloadrem:
          lda #228
          sec
          jsr MACPTR
  pla
  sta 1
    lda r10L
    sta sector_lba
    lda r10H
    sta sector_lba+1
    lda r11L
    sta sector_lba+2
    lda r11H
    sta sector_lba+3
rts


; String data
STR_HELLO_WORLD        .text "HELLO WORLD!",0
STR_INCOMPATIBLE_ROM   .text "ROM INCOMPATIBLE, REQUIRES R48 ",0
STR_ROM_OK             .text "ROM R48, OK ",0
